from collections import OrderedDict
from typing import Any
from typing import Dict


def convert_raw_k8s_cluster_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    """
    Util function to convert raw resource state from AWS K8s Cluster to present input format.
    """

    resource_id = raw_resource.get("id")
    resource_parameters = OrderedDict(
        {
            "name": "name",
            "controllerClusterId": "controller_cluster_id",
            "region": "region",
            "capacity": "capacity",
            "compute": "compute",
            "logging": "logging",
            "scheduling": "scheduling",
            "security": "security",
            "strategy": "strategy",
            "autoScaler": "auto_scaler",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    return resource_translated


def convert_raw_launch_spec_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    """
    Util function to convert raw resource state from AWS Launch Spec to present input format.
    """

    resource_id = raw_resource.get("id")
    resource_parameters = OrderedDict(
        {
            "name": "name",
            "taints": "taints",
            "tags": "tags",
            "subnetIds": "subnet_ids",
            "userData": "user_data",
            "securityGroupIds": "security_group_ids",
            "rootVolumeSize": "root_volume_size",
            "strategy": "strategy",
            "restrictScaleDown": "restrict_scale_down",
            "resourceLimits": "resource_limits",
            "oceanId": "ocean_id",
            "labels": "labels",
            "instanceTypes": "instance_types",
            "imageId": "image_id",
            "iamInstanceProfile": "iam_instance_profile",
            "blockDeviceMappings": "block_device_mappings",
            "elasticIpPool": "elastic_ip_pool",
            "preferredSpotTypes": "preferred_spot_types",
            "autoScale": "auto_scale",
            "associatePublicIpAddress": "associateP_public_ip_address",
            "scheduling": "scheduling",
            "instanceMetadataOptions": "instance_metadata_options",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated



import json
from typing import Any
from typing import Dict


async def get_active_roll(hub, ctx, ocean_cluster_id) -> Dict[str, Any]:
    result = {}
    url = f"{hub.exec.spotinst.URL}/ocean/aws/k8s/cluster/{ocean_cluster_id}/roll?accountId={ctx.acct.account_id}"
    ret = await hub.exec.request.json.get(
        ctx,
        url=url,
        success_codes=[200],
        headers={"Authorization": f"Bearer {ctx.acct.token}"},
    )

    if not ret["result"]:
        raise ValueError(
            f"Error on requesting GET {url} with status code {ret['status']}:"
            f" {ret.get('comment', '')}"
        )

    cluster_roll_list = ret.get("ret").get("response").get("items", {})
    for cluster_roll_item in cluster_roll_list:
        if (
            cluster_roll_item.get("scope") == "Cluster"
            and cluster_roll_item.get("status") == "IN_PROGRESS"
        ):
            result = cluster_roll_item

    return result


async def initiate_roll(hub, ctx, update_policy, ocean_cluster_id):
    roll_result = dict(comment=(), result=True, ret=None)
    if update_policy and update_policy["shouldRoll"]:
        active_roll = (
            await hub.tool.spotinst.ocean.aws.k8s_cluster_utils.get_active_roll(
                ctx, ocean_cluster_id
            )
        )

        if not active_roll:
            if ctx.get("test", False):
                roll_result["comment"] = (
                    f"Would initiate roll for ocean cluster {ocean_cluster_id}",
                )
                return roll_result

            roll_cofig = update_policy["roll"]
            ocean_custer_initial_roll_url = f"{hub.exec.spotinst.URL}/ocean/aws/k8s/cluster/{ocean_cluster_id}/roll?accountId={ctx.acct.account_id}"
            roll_data = {"roll": roll_cofig}

            ret = await hub.exec.request.json.post(
                ctx,
                success_codes=[200],
                headers={"Authorization": f"Bearer {ctx.acct.token}"},
                url=ocean_custer_initial_roll_url,
                data=json.dumps(roll_data),
            )
            if not ret["result"]:
                errors = json.loads(ret["ret"]).get("response").get("errors")
                if errors:
                    roll_result["comment"] = (
                        ret["comment"],
                        errors[0].get("message"),
                    )
                else:
                    roll_result["comment"] = (ret["comment"],)
                roll_result["result"] = False
                return roll_result

            roll_result["comment"] = (
                f"Initiated roll for ocean cluster {ocean_cluster_id}",
            )
        else:
            roll_result["comment"] = (
                f"Active roll exist for ocean cluster {ocean_cluster_id}",
            )

        return roll_result



from typing import Any
from typing import Dict


def handle_null(hub, desire_state: Dict[str, Any], current_state: Dict[str, Any]):
    """
    Assign current_state parameter value to desire_state parameter if desire_state parameter value is None
    or parameter key is missing in desire_state.
    """
    if isinstance(current_state, dict):
        for key, value in current_state.items():
            if key in desire_state:
                desire_value = desire_state.get(key)
                if desire_value is None:
                    desire_state[key] = value
                elif isinstance(desire_value, dict):
                    desire_state[key] = hub.tool.spotinst.state_utils.handle_null(
                        desire_value, value
                    )
            else:
                desire_state[key] = value

    return desire_state



def __init__(hub):
    # Provides the ctx argument to all execution modules
    # which will have profile info from the account module
    hub.exec.spotinst.ACCT = ["spotinst"]
    hub.exec.spotinst.URL = "https://api.spotinst.io"



async def gather(hub, profiles):
    """
    load profiles from unencrypted spotinst credential files

    Example:
    .. code-block:: yaml

        spotinst:

          default:
            account_id: act-d60b6ecd
            token: XXXXXXXX
          test_development_idem_spotinst:
            use_ssl: True
            account_id: act-d60b6ecd
            token: XXXXXXXXXXXX

    """
    sub_profiles = {}
    for profile, ctx in profiles.get("spotinst", {}).items():
        sub_profiles[profile] = ctx

    return sub_profiles



import collections.abc as abc
import copy
from typing import Any
from typing import Dict
from typing import List


def are_lists_identical(hub, list1: List, list2: List) -> bool:
    """
    Compare two lists and logs the difference.
    :param list1: first list.
    :param list2: second list.
    :return: true if there is no difference between both lists.
    :raises exception if one of the list  is not of type list
    """
    if (list1 is None or len(list1) == 0) and (list2 is None or len(list2) == 0):
        return True
    if list1 is None or len(list1) == 0 or list2 is None or len(list2) == 0:
        return False

    for l in [list1, list2]:
        if not isinstance(l, List):
            raise TypeError(
                f"Expecting lists to compare. This is expected to be of type List: '{l}'"
            )

    diff = [i for i in list1 + list2 if i not in list1 or i not in list2]
    result = len(diff) == 0
    if not result:
        hub.log.debug(f"There are {len(diff)} differences:\n{diff[:5]}")
    return result


def deep_diff(
    hub, current_state: Dict, desire_state: Dict, ignore: List = None
) -> Dict[str, Any]:
    """
    Recursively compare two dictionaries(current_state and desire_state) and find the differences.
    """
    ignore = ignore or []
    res = {}
    current_state = copy.deepcopy(current_state) or {}
    desire_state = copy.deepcopy(desire_state) or {}

    stack = [(current_state, desire_state, False)]

    while len(stack) > 0:
        tmps = []
        tmp_current_state, tmp_desire_state, reentrant = stack.pop()
        for key in set(list(tmp_current_state) + list(tmp_desire_state)):
            if (
                key in tmp_current_state
                and key in tmp_desire_state
                and isinstance(tmp_current_state[key], List)
                and hub.tool.spotinst.state_comparison_utils.are_lists_identical(
                    tmp_current_state[key], tmp_desire_state[key]
                )
            ):
                del tmp_current_state[key]
                del tmp_desire_state[key]
                continue
            elif (
                key in tmp_current_state
                and key in tmp_desire_state
                and tmp_current_state[key] == tmp_desire_state[key]
            ):
                del tmp_current_state[key]
                del tmp_desire_state[key]
                continue
            if not reentrant:
                if key in tmp_current_state and key in ignore:
                    del tmp_current_state[key]
                if key in tmp_desire_state and key in ignore:
                    del tmp_desire_state[key]
                if isinstance(tmp_current_state.get(key), abc.Mapping) and isinstance(
                    tmp_desire_state.get(key), abc.Mapping
                ):
                    tmps.append((tmp_current_state[key], tmp_desire_state[key], False))
        if tmps:
            stack.extend([(tmp_current_state, tmp_desire_state, True)] + tmps)
    if current_state:
        res["current_state"] = current_state
    if desire_state:
        res["desire_state"] = desire_state
    return res



from typing import List


def is_update_required(hub, old_tags_list: List, new_tags_list: List) -> bool:
    old_tags_list = old_tags_list if old_tags_list is not None else []
    new_tags_list = new_tags_list if new_tags_list is not None else []

    tags_to_add = {}
    tags_to_remove = {}
    new_tags = {}
    for tag in new_tags_list:
        new_tags[tag["tagKey"]] = tag.get("tagValue")

    old_tags = {}
    for tag in old_tags_list:
        old_tags[tag["tagKey"]] = tag.get("tagValue")

    for key, value in old_tags.items():
        if key in new_tags:
            if old_tags[key] != new_tags[key]:
                tags_to_remove.update({key: old_tags[key]})
                tags_to_add.update({key: new_tags[key]})
            new_tags.pop(key)
        else:
            tags_to_remove.update({key: old_tags[key]})
    tags_to_add.update(new_tags)

    return tags_to_add or tags_to_remove



import base64
import copy
import json
from typing import Any
from typing import Dict
from typing import List

import dict_tools.differ as differ

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    image_id: str,
    ocean_id: str,
    resource_id: str = None,
    iam_instance_profile: Dict = None,
    taints: List = None,
    tags: List = None,
    subnet_ids: List = None,
    user_data: str = None,
    security_group_ids: List = None,
    root_volume_size: int = None,
    strategy: Dict = None,
    restrict_scale_down: bool = None,
    resource_limits: Dict = None,
    labels: List = None,
    instance_types: List = None,
    preferred_spot_types: List = None,
    elastic_ip_pool: Dict = None,
    auto_scale: Dict = None,
    block_device_mappings: List = None,
    associate_public_ip_address: bool = None,
    scheduling: Dict = None,
    instance_metadata_options: Dict = None,
) -> Dict[str, Any]:
    r"""

     Create a Virtual Node Group.

     Please refer the `Spot Create Launch Spec(Virtual Node Group) documentation <https://docs.spot.io/api/#operation/OceanAWSLaunchSpecCreate>`_ to get insight of functionality and input parameters

     Args:
         name(Text): launch specification name.
         image_id(Text): AWS image identifier.
         ocean_id(Text): The Ocean cluster identifier. Required for Launch Spec creation
         resource_id(Text, optional): The virtual node group identifier.
         iam_instance_profile(Dict, optional): The instance profile iamRole object.
         taints(List, optional): Add taints to Launch Spec.
         tags(List, optional): List of kay-value pairs of tags.
         subnet_ids(List, optional): Set subnets in launchSpec. Each element in the array is a subnet identifier.
         user_data(Text, optional): Set user data script.
         security_group_ids(List, optional): Set security groups. Each element in the array is a security group identifier.
         root_volume_size(Integer, optional): Set root volume size (in GB). This field and blockDeviceMappings cannot be used in the same specification.
         strategy(Dict, optional): Similar to a strategy for an Ocean cluster, but applying only to a virtual node group.
         restrict_scale_down(Boolean, optional): When set to “True”, VNG nodes will be treated as if all pods running have the restrict-scale-down label. Therefore, Ocean will not scale nodes down unless empty.
         resource_limits(Dict, optional): Option to set a maximum/minimum number of instances per launch specification.
         labels(List, optional): An array of labels to add to the VNG nodes. Only custom user labels are allowed, and not Kubernetes built-in labels or Spot internal labels.
         instance_types(List, optional): A list of instance types allowed to be provisioned for pods pending for the launch specification.
         preferred_spot_types(List, optional): When Ocean scales up instances, it takes your preferred types into consideration while maintaining a variety of machine types running for optimized distribution.
         elastic_ip_pool(Dict, optional): Assign an Elastic IP to the instances launched by the launch spec.
         auto_scale(Dict, optional): Object specifying the automatic scaling of an Ocean VNG.
         block_device_mappings(List, optional):
            Block devices that are exposed to the instance. You can specify virtual devices and EBS volumes.
            This parameter and rootVolumeSize cannot be in the spec at the same time.
            This parameter can be null, but if not null, it must contain at least one block device.
         associate_public_ip_address(Boolean, optional): Configure public IP address allocation.
         scheduling(Dict, optional): An object used to define scheduled tasks such as a manual headroom update.
         instance_metadata_options(Dict, optional): Ocean instance metadata options object for IMDSv2.

    Request Syntax:
        [spotinst.ocean.aws.launch_spec-resource-id]:
             spotinst.ocean.aws.launch_spec.present:
               - name: 'string'
               - resource_id: 'string'
               - taints:
                 - effect: 'string'
                   key: 'string'
                   value: 'string'
               - tags:
                 - tagKey: 'string'
                   tagValue: 'string'
                 - tagKey: 'string'
                   tagValue: 'string'
               - subnet_ids:
                 - 'string'
               - user_data: 'string'
               - security_group_ids:
                 - 'string'
               - root_volume_size: Integer
               - resource_limits:
                   maxInstanceCount: Integer
               - ocean_id: 'string'
               - labels:
                 - key: 'string'
                   value: 'string'
                 - key: 'string'
                   value: 'string'
               - instance_types:
                 - 'string'
               - image_id: 'string'
               - iam_instance_profile:
                   arn: 'string'
               - associate_public_ip_address: true/false
               - preferred_spot_types:
                 - 'string'
               - elastic_ip_pool:
                   tagSelector:
                     tagKey: 'string'
                     tagValue: 'string'
               - auto_scale:
                   headrooms:
                    - cpuPerUnit: Integer
                      memoryPerUnit: Integer
                      numOfUnits: Integer
               - block_device_mappings:
                - deviceName: 'string'
                  ebs:
                    deleteOnTermination: true/false
                    encrypted: true/false
                    volumeSize: Integer
                    volumeType: 'string'
                - scheduling:
                    tasks:
                      - isEnabled: true/false
                        cronExpression: 'string'
                        taskType: 'string'
                    shutdownHours:
                      timeWindows:
                        - 'string'
                    isEnabled: true/false
                - instanceMetadataOptions:
                            httpTokens: 'string'
                            httpPutResponseHopLimit: Integer

     Returns:
         Dict[str, Any]

     Examples:

         .. code-block:: sls

             ols-77e27034:
               spotinst.ocean.aws.launch_spec.present:
               - name: cluster-arm
               - resource_id: ols-77e27034
               - taints:
                 - effect: NoSchedule
                   key: dedicated
                   value: idem-test-value
               - tags:
                 - tagKey: tag-key-1
                   tagValue: tag-value-1
                 - tagKey: tag-key-2
                   tagValue: tag-value-2
               - subnet_ids:
                 - subnet-111122223333abcd
                 - subnet-444455556666abcd
               - user_data: /etc/eks/idem-test.sh
               - security_group_ids:
                 - sg-111122223333abcd
                 - sg-444455556666abcd
               - root_volume_size: 8
               - resource_limits:
                   maxInstanceCount: 1
               - ocean_id: o-b78b1d69
               - labels:
                 - key: autoscale
                   value: 'false'
                 - key: dedicated
                   value: idem-test-value
               - instance_types:
                 - t2.micro
               - image_id: ami-111122223333abcd
               - iam_instance_profile:
                   arn: arn:aws:iam::111122223333:instance-profile/spotinst-eks-stack-NodeInstanceProfile
               - auto_scale:
                   headrooms:
                    - cpuPerUnit: 1024
                      memoryPerUnit: 512
                      numOfUnits: 3
               - block_device_mappings:
                - deviceName: /dev/xvda
                  ebs:
                    deleteOnTermination: true
                    encrypted: true
                    volumeSize: 30
                    volumeType: gp2
                - scheduling:
                    tasks:
                      - isEnabled: true
                        cronExpression: '0 1 * * *'
                        taskType: manualHeadroomUpdate
                    shutdownHours:
                      timeWindows:
                        - "Sat:08:00-Sun:08:00"
                    isEnabled: true
                - instanceMetadataOptions:
                            httpTokens: optional
                            httpPutResponseHopLimit: 12
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    user_data_base64 = None
    if user_data:
        try:
            user_data_base64 = base64.b64encode(user_data.encode("utf-8")).decode(
                "utf-8"
            )
        except UnicodeError as e:
            hub.log.debug(f"base 64 encoding failed for user_data {e}")
            result["result"] = False
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
            return result

    data = {
        "launchSpec": {
            "name": name,
            "taints": taints,
            "tags": tags,
            "subnetIds": subnet_ids,
            "userData": user_data_base64,
            "securityGroupIds": security_group_ids,
            "strategy": strategy,
            "restrictScaleDown": restrict_scale_down,
            "resourceLimits": resource_limits,
            "labels": labels,
            "instanceTypes": instance_types,
            "imageId": image_id,
            "iamInstanceProfile": iam_instance_profile,
            "elasticIpPool": elastic_ip_pool,
            "preferredSpotTypes": preferred_spot_types,
            "autoScale": auto_scale,
            "associatePublicIpAddress": associate_public_ip_address,
            "scheduling": scheduling,
            "instanceMetadataOptions": instance_metadata_options,
        }
    }

    # blockDeviceMappings and rootVolumeSize cannot be in the spec at the same time
    if block_device_mappings:
        data["launchSpec"]["blockDeviceMappings"] = block_device_mappings
    else:
        data["launchSpec"]["rootVolumeSize"] = root_volume_size

    resource_parameters = {
        "taints": taints,
        "tags": tags,
        "subnet_ids": subnet_ids,
        "user_data": user_data_base64,
        "security_group_ids": security_group_ids,
        "root_volume_size": root_volume_size,
        "strategy": strategy,
        "restrict_scale_down": restrict_scale_down,
        "resource_limits": resource_limits,
        "ocean_id": ocean_id,
        "labels": labels,
        "instance_types": instance_types,
        "image_id": image_id,
        "iam_instance_profile": iam_instance_profile,
        "elastic_ip_pool": elastic_ip_pool,
        "preferred_spot_types": preferred_spot_types,
        "auto_scale": auto_scale,
        "block_device_mappings": block_device_mappings,
        "associate_public_ip_address": associate_public_ip_address,
        "scheduling": scheduling,
        "instance_metadata_options": instance_metadata_options,
    }

    desire_state = {"name": name, "resource_id": resource_id}
    for parameter_key, parameter_value in resource_parameters.items():
        if parameter_value is not None:
            desire_state[parameter_key] = parameter_value

    before = None
    if resource_id:
        get_url = f"{hub.exec.spotinst.URL}/ocean/aws/k8s/launchSpec/{resource_id}?accountId={ctx.acct.account_id}"
        ret = await hub.exec.request.json.get(
            ctx,
            url=get_url,
            success_codes=[200],
            headers={"Authorization": f"Bearer {ctx.acct.token}"},
        )
        if not ret["result"]:
            if ret["status"] == 400:
                result["comment"] = (
                    f"spotinst.ocean.aws.launch_spec '{name}'(resource Id: '{resource_id}') not found",
                )  # In case of non-existance resource id, comment returns ' Bad Request'
            else:
                result["comment"] = ret["comment"]
            result["result"] = ret["result"]
            return result

        if ret["ret"].get("response").get("items"):
            before = ret["ret"].get("response").get("items")[0]

    if before:
        result[
            "old_state"
        ] = hub.tool.spotinst.ocean.aws.conversion_utils.convert_raw_launch_spec_to_present(
            raw_resource=before,
            idem_resource_name=resource_id,
        )
        desire_state = hub.tool.spotinst.state_utils.handle_null(
            desire_state=desire_state, current_state=result["old_state"]
        )
        diff = hub.tool.spotinst.state_comparison_utils.deep_diff(
            result["old_state"].copy(), desire_state.copy(), ignore="tags"
        )
        update_required = False
        if diff and "desire_state" in diff:
            for item in diff["desire_state"]:
                if diff["desire_state"].get(item):  # if value is not None
                    update_required = True

        is_tags_updated = hub.tool.spotinst.ocean.aws.tag_utils.is_update_required(
            old_tags_list=result["old_state"].get("tags"),
            new_tags_list=desire_state.get("tags"),
        )
        update_required = update_required or is_tags_updated

        if update_required:
            if ctx.get("test", False):
                result["new_state"] = desire_state
                result[
                    "comment"
                ] = hub.tool.spotinst.comment_utils.would_update_comment(
                    resource_type="spotinst.ocean.aws.launch_spec", name=resource_id
                )
                return result
            # Update in real
            try:
                ret = await hub.exec.request.json.put(
                    ctx,
                    success_codes=[200],
                    headers={"Authorization": f"Bearer {ctx.acct.token}"},
                    url=f"{hub.exec.spotinst.URL}/ocean/aws/k8s/launchSpec/{resource_id}?accountId={ctx.acct.account_id}",
                    data=json.dumps(data),
                )
            except Exception as e:
                hub.log.debug(f"Could not update spotinst.ocean.aws.launch_spec {e}")
                result["result"] = False
                result["comment"] = (f"{e.__class__.__name__}: {e}",)
                return result

            if not ret["result"]:
                errors = json.loads(ret["ret"]).get("response").get("errors")
                if errors:
                    result["comment"] = (
                        ret["comment"],
                        errors[0].get("message"),
                    )
                else:
                    result["comment"] = ret["comment"]
                result["result"] = False
                return result

            result[
                "new_state"
            ] = hub.tool.spotinst.ocean.aws.conversion_utils.convert_raw_launch_spec_to_present(
                raw_resource=ret["ret"].get("response").get("items")[0],
                idem_resource_name=name,
            )
            result[
                "comment"
            ] = f"Updated spotinst.ocean.aws.launch_spec Id: '{resource_id}')"
        else:
            result["comment"] = hub.tool.spotinst.comment_utils.already_present_comment(
                resource_type="spotinst.ocean.aws.launch_spec",
                name=name,
                resource_id=resource_id,
            )
            result["new_state"] = copy.deepcopy(result["old_state"])
            return result
    else:  # create new resource
        if ctx.get("test", False):
            # For test we should add resource_id
            if desire_state.get("resource_id", None) is None:
                desire_state["resource_id"] = "resource_id_known_after_present"
            result["new_state"] = desire_state
            result["comment"] = hub.tool.spotinst.comment_utils.would_create_comment(
                resource_type="spotinst.ocean.aws.launch_spec", name=name
            )
            return result
        # create in real
        data.get("launchSpec")["oceanId"] = ocean_id
        try:
            ret = await hub.exec.request.json.post(
                ctx,
                success_codes=[200],
                headers={"Authorization": f"Bearer {ctx.acct.token}"},
                url=f"{hub.exec.spotinst.URL}/ocean/aws/k8s/launchSpec?accountId={ctx.acct.account_id}",
                data=json.dumps(data),
            )
        except Exception as e:
            hub.log.debug(f"Could not create spotinst.ocean.aws.launch_spec {e}")
            result["result"] = False
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
            return result

        if not ret["result"]:
            errors = json.loads(ret["ret"]).get("response").get("errors")
            if errors:
                result["comment"] = (
                    ret["comment"],
                    errors[0].get("message"),
                )
            else:
                result["comment"] = ret["comment"]
            result["result"] = False
            return result

        result[
            "new_state"
        ] = hub.tool.spotinst.ocean.aws.conversion_utils.convert_raw_launch_spec_to_present(
            raw_resource=ret["ret"].get("response").get("items")[0],
            idem_resource_name=name,
        )

        result["comment"] = hub.tool.spotinst.comment_utils.create_comment(
            resource_type="spotinst.ocean.aws.launch_spec", name=name
        )

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    Deletes the specified launch Spec.
    When this call completes, the launch configuration is no longer available for use.

    Please refer the `Spot Delete Launch Spec(Virtual Node Group) documentation <https://docs.spot.io/api/#operation/OceanAWSLaunchSpecDelete>`_ to get insight of functionality and input parameters


    Args:
        name(Text): An idem name of the launch spec.
        resource_id(Text): The AWS ID of the launch spec.

    Request Syntax:
       [launch_spec-resource-id]:
              spotinst.ocean.aws.launch_spec.absent:
              - name: 'string'
              - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem-test-launch_spec:
              spotinst.ocean.aws.launch_spec.absent:
                - name: idem-test-launch_spec
                - resource_id:idem-test-launch_spec
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.spotinst.comment_utils.already_absent_comment(
            resource_type="spotinst.ocean.aws.launch_spec", name=name
        )
        return result

    url = f"{hub.exec.spotinst.URL}/ocean/aws/k8s/launchSpec/{resource_id}?accountId={ctx.acct.account_id}"
    before = None
    try:
        before = await hub.exec.request.json.get(
            ctx,
            url=url,
            success_codes=[200],
            headers={"Authorization": f"Bearer {ctx.acct.token}"},
        )
    except Exception as e:
        hub.log.debug(f"Could not get Launch Spec {e}")
        result["result"] = False
        result["comment"] = (f"{e.__class__.__name__}: {e}",)
        return result

    if before["status"] == 400:
        result["comment"] = hub.tool.spotinst.comment_utils.already_absent_comment(
            resource_type="spotinst.ocean.aws.launch_spec", name=name
        )
    elif before["result"]:
        result[
            "old_state"
        ] = hub.tool.spotinst.ocean.aws.conversion_utils.convert_raw_launch_spec_to_present(
            raw_resource=before["ret"].get("response").get("items")[0],
            idem_resource_name=name,
        )

        if ctx.get("test", False):
            result["comment"] = hub.tool.spotinst.comment_utils.would_delete_comment(
                resource_type="spotinst.ocean.aws.launch_spec", name=name
            )
            return result
        try:
            ret = await hub.exec.request.json.delete(
                ctx,
                url=url,
                success_codes=[200],
                headers={"Authorization": f"Bearer {ctx.acct.token}"},
            )
            result["result"] = ret["result"]
            if not result["result"]:
                hub.log.debug(
                    f"Could not delete spotinst.ocean.aws.launch_spec {result['comment']} {result['ret']}"
                )
                result["comment"] = ret["comment"]
                return result

            result["comment"] = hub.tool.spotinst.comment_utils.delete_comment(
                resource_type="spotinst.ocean.aws.launch_spec", name=name
            )
        except Exception as e:
            hub.log.debug(f"Could not delete spotinst.ocean.aws.launch_spec {e}")
            result["result"] = False
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
            return result
    else:
        hub.log.debug(f"Could not get Launch Spec {before['comment']} {before['ret']}")
        result["result"] = False
        result["comment"] = before["comment"]

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function
    Gets information about the virtual node groups for the cluster.

    Please refer the `Spot Listing Launch Spec(Virtual Node Group) documentation <https://docs.spot.io/api/#operation/OceanAWSLaunchSpecList>`_ to get insight of functionality and input parameters

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe spotinst.ocean.aws.launch_spec
    """

    result = {}

    url = f"{hub.exec.spotinst.URL}/ocean/aws/k8s/launchSpec?accountId={ctx.acct.account_id}"
    ret = await hub.exec.request.json.get(
        ctx,
        url=url,
        success_codes=[200],
        headers={"Authorization": f"Bearer {ctx.acct.token}"},
    )

    if not ret["result"]:
        hub.log.debug(f"Could not describe launch_spec {ret['comment']}")
        return {}

    for launch_spec_item in ret["ret"].get("response").get("items"):
        resource_id = launch_spec_item.get("id")
        resource_translated = hub.tool.spotinst.ocean.aws.conversion_utils.convert_raw_launch_spec_to_present(
            raw_resource=launch_spec_item, idem_resource_name=resource_id
        )
        result[resource_id] = {
            "spotinst.ocean.aws.launch_spec.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



import base64
import copy
import json
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    controller_cluster_id: str,
    region: str,
    compute: Dict = None,
    resource_id: str = None,
    capacity: Dict = None,
    logging: Dict = None,
    scheduling: Dict = None,
    security: Dict = None,
    strategy: Dict = None,
    auto_scaler=None,
    update_policy: Dict = None,
) -> Dict[str, Any]:
    r"""

    Create a Ocean k8s Cluster.

    Please refer the `Spot Create Ocean Cluster documentation <https://docs.spot.io/api/#operation/OceanAWSClusterCreate>`_ to get insight of functionality and input parameters
    Args:

        name(Text): ocean cluster name.
        resource_id(Text, optional): The ocean cluster identifier.
        auto_scaler(Dict, optional): The automatic scaling mechanism used in Ocean for Kubernetes.
        capacity (Dict, optional): The overall capability of the Ocean cluster expressed as number of instances and specified with a minimum,
                                   a maximum, and a target number of running instances.
        compute (Dict, optional):
            Compute specifications for the Ocean cluster. This is required for greenfield and optional for brownfield.
        controller_cluster_id(string, required): Reporting identifier for the Ocean Controller.
        logging	(Dict, optional):  Logging configuration for ocean aws cluster.
        region(string, required): Region for the Ocean cluster to run in.
        scheduling(Dict, optional)	: An object used to define times for a task such as a shutdown to be activated.
        security(Dict, optional): Object for cluster security features.
        strategy(Dict, optional): An object defining the cluster strategy with regard to waiting periods and utilization of on-demand and reserved instances.
        update_policy(Dict, optional): Configures the cluster update policy. You can configure to start the roll on update and run it in defined batches.

    Request Syntax:
       [spotinst.ocean.aws.k8s_cluster-state-id]:
             spotinst.ocean.aws.k8s_cluster.present:
              - name: 'string'
              - controller_cluster_id: 'string'
              - region: 'string'
              - update_policy:
                  shouldRoll: boolean
                  roll:
                     batchSizePercentage: Integer
                     comment: 'string'
                     batchMinHealthyPercentage: Integer
              - compute:
                  launchSpecification:
                    keyPair: 'string'
                    imageId: 'string'
                    associatePublicIpAddress: boolean
                    rootVolumeSize: Integer
                    iamInstanceProfile:
                        arn: 'string'
                    userData: 'string'
                    tags:
                    - tagKey: 'string'
                      tagValue: 'string'
                    - tagKey: 'string'
                      tagValue: 'string'
                    securityGroupIds:
                        - 'string'
                  subnetIds:
                     - 'string'
                  instanceTypes:
                    whitelist:
                    - 'string'
              - strategy:
                  fallbackToOd: boolean
                  spotPercentage: Integer
                  utilizeReservedInstances: boolean
              - auto_scaler:
                  isEnabled: boolean
                  isAutoConfig: boolean
                  cooldown: Integer
                  headroom:
                    cpuPerUnit: Integer
                    memoryPerUnit: Integer
                    numOfUnits: Integer
                  down:
                    maxScaleDownPercentage: Integer
              - resourceLimits:
                  maxMemoryGib: Integer
                  maxVCpu: Integer
              - capacity:
                  maximum: Integer
                  minimum: Integer

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            ocean_cluster:
              spotinst.ocean.aws.k8s_cluster.present:
              - name: idem-test
              - controller_cluster_id: idem-test
              - region: ap-east-1
              - update_policy:
                  shouldRoll: false
                  roll:
                     batchSizePercentage: 30
                     comment: Idem cluster roll testing
                     batchMinHealthyPercentage: 100
              - compute:
                  launchSpecification:
                    keyPair: test-idem-key
                    imageId: ami-06c9598a8dcd87e79
                    associatePublicIpAddress: False
                    rootVolumeSize: 100
                    iamInstanceProfile:
                        arn: arn:aws:iam::111222333:instance-profile/idem-instance-profile-test
                    userData: |
                            #!/bin/bash -xe
                            /etc/eks/test-idem.sh
                    tags:
                    - tagKey: tag-ket-1
                      tagValue: tag-value-1
                    - tagKey: tag-ket-2
                      tagValue: tag-value-2
                    securityGroupIds:
                        - sg-qbc111122223333
                  subnetIds:
                     - subnet-0abc111222333
                  instanceTypes:
                    whitelist:
                    - c5.large
              - strategy:
                  fallbackToOd: true
                  spotPercentage: 100
                  utilizeReservedInstances: true
              - auto_scaler:
                  isEnabled: True
                  isAutoConfig: true
                  cooldown: 300
                  headroom:
                    cpuPerUnit: 1024
                    memoryPerUnit: 512
                    numOfUnits: 3
                  down:
                    maxScaleDownPercentage: 60
                  resourceLimits:
                    maxMemoryGib: 20000
                    maxVCpu: 100000
              - capacity:
                  maximum: 1000
                  minimum: 1

    """

    if auto_scaler is None:
        auto_scaler = {}
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # If userdata is set in compute, encode userData to base64
    if (
        compute
        and compute.get("launchSpecification")
        and compute.get("launchSpecification").get("userData")
    ):
        try:
            user_data_bytes = (
                compute.get("launchSpecification").get("userData").encode("utf-8")
            )
            base64_string = base64.b64encode(user_data_bytes).decode("utf-8")
            compute["launchSpecification"]["userData"] = base64_string
        except UnicodeError as e:
            hub.log.debug(f"base 64 encoding failed for user_data {e}")
            result["result"] = False
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
            return result

    data = {
        "cluster": {
            "name": name,
            "capacity": capacity,
            "compute": compute,
            "controllerClusterId": controller_cluster_id,
            "logging": logging,
            "scheduling": scheduling,
            "security": security,
            "strategy": strategy,
        }
    }

    if auto_scaler:
        data["cluster"]["autoScaler"] = auto_scaler

    resource_parameters = {
        "capacity": capacity,
        "compute": compute,
        "controller_cluster_id": controller_cluster_id,
        "logging": logging,
        "region": region,
        "scheduling": scheduling,
        "security": security,
        "strategy": strategy,
        "auto_scaler": auto_scaler,
    }

    desire_state = {"name": name, "resource_id": resource_id}
    for parameter_key, parameter_value in resource_parameters.items():
        if parameter_value is not None:
            desire_state[parameter_key] = parameter_value

    before = None

    if resource_id:
        ocean_custer_resource_url = f"{hub.exec.spotinst.URL}/ocean/aws/k8s/cluster/{resource_id}?accountId={ctx.acct.account_id}"
        ret = await hub.exec.request.json.get(
            ctx,
            url=ocean_custer_resource_url,
            success_codes=[200],
            headers={"Authorization": f"Bearer {ctx.acct.token}"},
        )
        if not ret["result"]:
            if ret["status"] == 400:
                result["comment"] = (
                    f"spotinst.ocean.aws.k8s_cluster '{name}'(resource Id: '{resource_id}') not found",
                )  # In case of non-existance resource id, comment returns ' Bad Request'
            else:
                result["comment"] = ret["comment"]
            result["result"] = ret["result"]
            return result

        if ret["ret"].get("response").get("items"):
            before = ret["ret"].get("response").get("items")[0]

    if before:
        result[
            "old_state"
        ] = hub.tool.spotinst.ocean.aws.conversion_utils.convert_raw_k8s_cluster_to_present(
            raw_resource=before,
            idem_resource_name=resource_id,
        )

        desire_state = hub.tool.spotinst.state_utils.handle_null(
            desire_state=desire_state, current_state=result["old_state"]
        )
        diff = hub.tool.spotinst.state_comparison_utils.deep_diff(
            result["old_state"].copy(), desire_state.copy(), ignore="tags"
        )

        update_required = False
        if diff and "desire_state" in diff:
            for item in diff["desire_state"]:
                if diff["desire_state"].get(item):  # if value is not None
                    update_required = True

        is_tags_updated = hub.tool.spotinst.ocean.aws.tag_utils.is_update_required(
            old_tags_list=result["old_state"]
            .get("compute")
            .get("launchSpecification")
            .get("tags"),
            new_tags_list=desire_state.get("compute")
            .get("launchSpecification")
            .get("tags"),
        )
        update_required = update_required or is_tags_updated

        if update_required:
            if ctx.get("test", False):
                result["new_state"] = desire_state
                result[
                    "comment"
                ] = hub.tool.spotinst.comment_utils.would_update_comment(
                    resource_type="spotinst.ocean.aws.k8s_cluster", name=resource_id
                )
                if update_policy and update_policy["shouldRoll"]:
                    roll_result = await hub.tool.spotinst.ocean.aws.k8s_cluster_utils.initiate_roll(
                        ctx, update_policy, resource_id
                    )
                    result["comment"] += roll_result["comment"]
                    result["new_state"]["update_policy"] = copy.deepcopy(update_policy)

                return result
            # update in real
            ret = await hub.exec.request.json.put(
                ctx,
                success_codes=[200],
                headers={"Authorization": f"Bearer {ctx.acct.token}"},
                url=ocean_custer_resource_url,
                data=json.dumps(data),
            )
            if not ret["result"]:
                errors = json.loads(ret["ret"]).get("response").get("errors")
                if errors:
                    result["comment"] = (
                        ret["comment"],
                        errors[0].get("message"),
                    )
                else:
                    result["comment"] = ret["comment"]
                result["result"] = False
                return result

            result[
                "new_state"
            ] = hub.tool.spotinst.ocean.aws.conversion_utils.convert_raw_k8s_cluster_to_present(
                raw_resource=ret["ret"].get("response").get("items")[0],
                idem_resource_name=resource_id,
            )
            result["comment"] = hub.tool.spotinst.comment_utils.update_comment(
                resource_type="spotinst.ocean.aws.k8s_cluster", name=resource_id
            )

            # initiate roll only in real update (when update is successful)
            if update_policy and update_policy["shouldRoll"]:
                roll_result = (
                    await hub.tool.spotinst.ocean.aws.k8s_cluster_utils.initiate_roll(
                        ctx, update_policy, resource_id
                    )
                )

                result["comment"] += roll_result["comment"]
                if not roll_result["result"]:
                    result["result"] = False
                    return result

                result["new_state"]["update_policy"] = copy.deepcopy(update_policy)
        else:
            result["comment"] = hub.tool.spotinst.comment_utils.already_present_comment(
                resource_type="spotinst.ocean.aws.k8s_cluster",
                name=name,
                resource_id=resource_id,
            )
            result["new_state"] = copy.deepcopy(result["old_state"])
            return result
    else:  # create new resource
        if ctx.get("test", False):
            # For test we should add resource_id
            if desire_state.get("resource_id", None) is None:
                desire_state["resource_id"] = "resource_id_known_after_present"
            result["new_state"] = desire_state
            result["comment"] = hub.tool.spotinst.comment_utils.would_create_comment(
                resource_type="spotinst.ocean.aws.k8s_cluster", name=name
            )
            return result
        # create in real
        try:
            data.get("cluster")["region"] = region
            ret = await hub.exec.request.json.post(
                ctx,
                success_codes=[200],
                headers={"Authorization": f"Bearer {ctx.acct.token}"},
                url=f"{hub.exec.spotinst.URL}/ocean/aws/k8s/cluster?accountId={ctx.acct.account_id}",
                data=json.dumps(data),
            )
        except Exception as e:
            hub.log.debug(f"Could not create spotinst.ocean.aws.k8s_cluster {e}")
            result["result"] = False
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
            return result

        if not ret["result"]:
            result["result"] = False
            errors = json.loads(ret["ret"]).get("response").get("errors")
            if errors:
                result["comment"] = (
                    ret["comment"],
                    errors[0].get("message"),
                )
            else:
                result["comment"] = ret["comment"]
            return result

        resource_id = ret["ret"].get("response").get("items")[0].get("id")
        result[
            "new_state"
        ] = hub.tool.spotinst.ocean.aws.conversion_utils.convert_raw_k8s_cluster_to_present(
            raw_resource=ret["ret"].get("response").get("items")[0],
            idem_resource_name=resource_id,
        )
        result["comment"] = hub.tool.spotinst.comment_utils.create_comment(
            resource_type="spotinst.ocean.aws.k8s_cluster", name=name
        )

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""

    Deletes the specified Ocean cluster.
    When this call completes, the cluster is no longer available for use.

    Please refer the `Spot Delete Ocean Cluster documentation <https://docs.spot.io/api/#operation/OceanAWSClusterDelete>`_ to get insight of functionality and input parameters

    Args:
        name(Text): An idem name of the ocean cluster.
        resource_id(Text): The ID of the ocean cluster.

    Request Syntax:
       [k8s_cluster-resource-id]:
              spotinst.ocean.aws.k8s_cluster.absent:
              - name: 'string'
              - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem-test-k8s_cluster:
              spotinst.ocean.aws.k8s_cluster.absent:
                - name: idem-test-k8s_cluster
                - resource_id:idem-test-k8s_cluster
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.spotinst.comment_utils.already_absent_comment(
            resource_type="spotinst.ocean.aws.k8s_cluster", name=name
        )
        return result

    url = f"{hub.exec.spotinst.URL}/ocean/aws/k8s/cluster/{resource_id}?accountId={ctx.acct.account_id}"
    try:
        before = await hub.exec.request.json.get(
            ctx,
            url=url,
            success_codes=[200],
            headers={"Authorization": f"Bearer {ctx.acct.token}"},
        )
    except Exception as e:
        hub.log.debug(f"Could not get K8s Cluster {e}")
        result["result"] = False
        result["comment"] = (f"{e.__class__.__name__}: {e}",)
        return result

    if before["status"] == 400:
        result["comment"] = hub.tool.spotinst.comment_utils.already_absent_comment(
            resource_type="spotinst.ocean.aws.k8s_cluster", name=name
        )
    elif before["result"]:
        result[
            "old_state"
        ] = hub.tool.spotinst.ocean.aws.conversion_utils.convert_raw_k8s_cluster_to_present(
            raw_resource=before["ret"].get("response").get("items")[0],
            idem_resource_name=resource_id,
        )

        if ctx.get("test", False):
            result["comment"] = hub.tool.spotinst.comment_utils.would_delete_comment(
                resource_type="spotinst.ocean.aws.k8s_cluster", name=name
            )
            return result
        try:
            after = await hub.exec.request.json.delete(
                ctx,
                url=url,
                success_codes=[200],
                headers={"Authorization": f"Bearer {ctx.acct.token}"},
            )
        except Exception as e:
            hub.log.debug(f"Could not delete spotinst.ocean.aws.k8s_cluster {e}")
            result["result"] = False
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
            return result
        if not after["result"] or after["status"] == 400:
            hub.log.debug(
                f"Could not delete spotinst.ocean.aws.k8s_cluster {result['comment']} {result['ret']}"
            )
            result["comment"] = after["comment"]
            return result

        result["result"] = after["result"]
        result["comment"] = hub.tool.spotinst.comment_utils.delete_comment(
            resource_type="spotinst.ocean.aws.k8s_cluster", name=name
        )
    else:
        hub.log.debug(f"Could not get k8s Cluster {before['comment']} {before['ret']}")
        result["result"] = False
        result["comment"] = before["comment"]

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function
    Gets information about the ocean clusters.

    Please refer the `Spot Listing Ocean Cluster documentation <https://docs.spot.io/api/#operation/OceanAWSClusterList>`_ to get insight of functionality and input parameters

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe spotinst.ocean.aws.k8s_cluster
    """

    result = {}
    url = (
        f"{hub.exec.spotinst.URL}/ocean/aws/k8s/cluster?accountId={ctx.acct.account_id}"
    )
    ret = await hub.exec.request.json.get(
        ctx,
        url=url,
        success_codes=[200],
        headers={"Authorization": f"Bearer {ctx.acct.token}"},
    )

    if not ret["result"]:
        raise ValueError(
            f"Error on requesting GET {url} with status code {ret['status']}:"
            f" {ret.get('comment', '')}"
        )
    cluster_list = ret.get("ret").get("response").get("items", None)
    for cluster_item in cluster_list:
        resource_id = cluster_item.get("id")
        translated_resource = hub.tool.spotinst.ocean.aws.conversion_utils.convert_raw_k8s_cluster_to_present(
            raw_resource=cluster_item, idem_resource_name=resource_id
        )

        result[resource_id] = {
            "spotinst.ocean.aws.k8s_cluster.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }
    return result
